# Generated by Django 2.2.5 on 2019-09-23 15:12

from django.db import migrations


def forward(apps, schema_editor):
    MicrosoftAccount = apps.get_model('sites_microsoft_auth', 'MicrosoftAccount')
    XboxLiveAccount = apps.get_model('sites_microsoft_auth', 'XboxLiveAccount')
    for account in MicrosoftAccount.objects.all():
        account.site = account.user.site
        account.save()
    for account in XboxLiveAccount.objects.all():
        account.site = account.user.site
        account.save()


class Migration(migrations.Migration):

    dependencies = [
        ('sites_microsoft_auth', '0003_auto_20190923_1511'),
    ]

    operations = [
        migrations.RunPython(forward, migrations.RunPython.noop)
    ]



from importlib import import_module

from django.test.signals import setting_changed
from django.utils.functional import SimpleLazyObject
from django.utils.translation import ugettext_lazy as _

constance_config = None
settings = None

""" List of all possible default configs for sites_microsoft_auth

    DEFAULT_CONFIG['defaults'] and DEFAULT_CONFIG['fieldsets'] are in a format
    usable by django-constance
    (https://django-constance.readthedocs.io/en/latest/) so these values can
    be directly added to the CONSTANCE_CONFIG and CONSTANCE_CONFIG_FIELDSETS
    in your global settings

    django-constance will also require the following field definition:

    CONSTANCE_ADDITIONAL_FIELDS = {
        'microsoft_choices': ['django.forms.fields.ChoiceField', {
            'widget': 'django.forms.Select',
            'choices': (('ma', 'Microsoft Accounts'),
                        ('xbl', 'Xbox Live Accounts'))}]}

    MICROSOFT_AUTH_CONFIG_CLASS is the only sites_microsoft_auth setting not present
    here. See bottom of file for more on it.
"""

LOGIN_TYPE_MA = "ma"
LOGIN_TYPE_XBL = "xbl"
HOOK_SETTINGS = [
    "MICROSOFT_AUTH_AUTHENTICATE_HOOK",
    "MICROSOFT_AUTH_CALLBACK_HOOK",
]
CACHE_TIMEOUT = 86400
CACHE_KEY_OPENID = "microsoft_auth_openid_config"
CACHE_KEY_JWKS = "microsoft_auth_jwks"

DEFAULT_CONFIG = {
    "defaults": {
        "MICROSOFT_AUTH_LOGIN_ENABLED": (
            True,
            _("Whether or not Microsoft OAuth login is enabled."),
            bool,
        ),
        "MICROSOFT_AUTH_LOGIN_TYPE": (
            LOGIN_TYPE_MA,
            _(
                """Type of Microsoft login to use.
                Microsoft Accounts is normal Microsoft login.
                Xbox Live Accounts use the old Microsoft Account login screen
                and then also authenticate against Xbox Live to retrieve
                Gamertag."""
            ),
            "microsoft_choices",
        ),
        "MICROSOFT_AUTH_TENANT_ID": (
            "common",
            _("Microsoft Office 365 Tenant ID"),
            str,
        ),
        "MICROSOFT_AUTH_CLIENT_ID": (
            "",
            _(
                """Microsoft OAuth Client ID, see
                https://apps.dev.microsoft.com/ for more."""
            ),
            str,
        ),
        "MICROSOFT_AUTH_CLIENT_SECRET": (
            "",
            _(
                """Microsoft OAuth Client Secret, see
                https://apps.dev.microsoft.com/ for more."""
            ),
            str,
        ),
        "MICROSOFT_AUTH_EXTRA_SCOPES": (
            "",
            _(
                """Extra OAuth scopes for authentication. Required
                scopes are always provided ('openid email'
                for Microsoft Auth and 'XboxLive.signin
                XboxLive.offline_access' for Xbox). Scopes are space
                delimited."""
            ),
            str,
        ),
        "MICROSOFT_AUTH_AUTO_CREATE": (
            True,
            _(
                """Autocreate user that attempt to login if they do not
                already exist?"""
            ),
            bool,
        ),
        "MICROSOFT_AUTH_REGISTER_INACTIVE_ADMIN": (
            False,
            _(
                """Automatically register admin class for auth type
                that is not active (Xbox when Microsoft Auth is
                enabled and Microsoft Auth when Xbox is enabled).
                Requires restart of app for setting to take effect."""
            ),
            bool,
        ),
        "MICROSOFT_AUTH_XBL_SYNC_USERNAME": (
            False,
            _(
                """Automatically sync the username from the Xbox Live
                Gamertag?"""
            ),
            bool,
        ),
        "MICROSOFT_AUTH_AUTO_REPLACE_ACCOUNTS": (
            False,
            _(
                """Automatically replace an existing Microsoft Account
                paired to a user when authenticating."""
            ),
            bool,
        ),
        "MICROSOFT_AUTH_AUTHENTICATE_HOOK": (
            "",
            _(
                """Callable hook to call after authenticating a user on the
                `sites_microsoft_auth.backends.MicrosoftAuthenticationBackend`.

                If the login type is Microsoft Auth, the parameters will be
                `(User: user, oauthlib.oauth2.rfc6749.tokens.OAuth2Token:
                token)`

                If the login type is Xbox Live, the parameters will be
                `(User:user, dict: token)` where token is the Xbox Token,
                see `sites_microsoft_auth.client.MicrosoftClient.fetch_xbox_token`
                for format"""
            ),
            str,
        ),
        "MICROSOFT_AUTH_CALLBACK_HOOK": (
            "",
            _(
                """Callable hook to call right before completing the `auth_callback` view.

                Really useful for adding custom data to message or chaning the
                expected base URL that gets passed back up to the window that
                initiated the original Authorize request.

                The parameters that will be passed will be `(HttpRequest:
                request, dict: context)`.

                The expected return value is the updated context dictionary.
                You should NOT remove the data that is currently there.

                `base_url` is the expected root URL of the window that
                initiated the authorize request

                `message` is a dictionary that will be serialized as a JSON
                string and passoed back to the initiating window.
                """
            ),
            str,
        ),
    },
    "fieldsets": {
        "Microsoft Login": (
            "MICROSOFT_AUTH_LOGIN_ENABLED",
            "MICROSOFT_AUTH_LOGIN_TYPE",
            "MICROSOFT_AUTH_TENANT_ID",
            "MICROSOFT_AUTH_CLIENT_ID",
            "MICROSOFT_AUTH_CLIENT_SECRET",
            "MICROSOFT_AUTH_EXTRA_SCOPES",
            "MICROSOFT_AUTH_AUTO_CREATE",
            "MICROSOFT_AUTH_REGISTER_INACTIVE_ADMIN",
            "MICROSOFT_AUTH_XBL_SYNC_USERNAME",
            "MICROSOFT_AUTH_AUTO_REPLACE_ACCOUNTS",
            "MICROSOFT_AUTH_AUTHENTICATE_HOOK",
            "MICROSOFT_AUTH_CALLBACK_HOOK",
        )
    },
    "fields": {
        "microsoft_choices": [
            "django.forms.fields.ChoiceField",
            {
                "widget": "django.forms.Select",
                "choices": (
                    (LOGIN_TYPE_MA, "Microsoft Auth"),
                    (LOGIN_TYPE_XBL, "Xbox Live"),
                ),
            },
        ]
    },
}


class SimpleConfig:
    def __init__(self, config=None):
        self._defaults = {}
        if config:
            self.add_default_config(config)

    def add_default_config(self, config):
        tmp_dict = {}
        for key, value in config["defaults"].items():
            tmp_dict[key] = value[0]

        self._defaults.update(tmp_dict)

    def __getattr__(self, attr):
        val = None

        # Django settings take priority
        try:
            # Check if present in user settings
            val = getattr(settings, attr)
        except AttributeError:
            pass

        # Check Constance first if it is installed
        if val is None and constance_config:
            try:
                val = getattr(constance_config, attr)
            except AttributeError:
                pass

        if val is None:
            # Fall back to defaults
            try:
                val = self._defaults[attr]
            except KeyError:
                raise AttributeError

        return val


def init_config():
    global config, constance_config, settings

    from django.conf import settings as django_settings

    settings = django_settings

    # set constance config global
    if "constance" in settings.INSTALLED_APPS:
        from constance import config as constance_config
    else:
        constance_config = None

    # retrieve and set config class
    if (
        hasattr(settings, "MICROSOFT_AUTH_CONFIG_CLASS")
        and settings.MICROSOFT_AUTH_CONFIG_CLASS is not None
    ):
        module, _, obj = settings.MICROSOFT_AUTH_CONFIG_CLASS.rpartition(".")
        conf = import_module(module)
        config = getattr(conf, obj)
        if hasattr(config, "add_default_config"):
            config.add_default_config(DEFAULT_CONFIG)
    else:
        config = SimpleConfig(DEFAULT_CONFIG)

    return config


""" Override MICROSOFT_AUTH_CONFIG_CLASS to inject your own custom dynamic
    settings class into sites_microsoft_auth. Useful if you want to manage config
    using a dynamic settings manager such as django-constance

    Optionally the class can have an 'add_default_config' method to add the
    above DEFAULT_CONFIG to config manager
"""
config = SimpleLazyObject(init_config)


def reload_settings(*args, **kwargs):
    global config

    setting = kwargs.get("setting", kwargs.get("key"))

    # only reinitialize config if settings changed
    if setting.startswith("MICROSOFT_AUTH_"):
        init_config()


setting_changed.connect(reload_settings)



import json
from unittest.mock import Mock, patch

from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.core.signing import TimestampSigner
from django.test import RequestFactory
from django.urls import reverse

from sites_microsoft_auth.views import AuthenticateCallbackView

from . import TestCase

STATE = TimestampSigner().sign(
    "e4675ea8d28a41b8b416fe9ed1fb52b1e4675ea8d28a41b8b416fe9ed1fb52b1"
)
EXPIRED_STATE = (
    "e4675ea8d28a41b8b416fe9ed1fb52b1e4675ea8d28a41b8b416fe9ed1fb52b1:"
    "1h5CgL:G-QiLZ3hetUPgrdpJlvAfXkZ2RQ"
)
TEST_ERROR = "test"
TEST_ERROR_DESCRIPTION = "some_error"


class ViewsTests(TestCase):
    def setUp(self):
        super().setUp()

        User = get_user_model()
        self.factory = RequestFactory()
        self.request = self.factory.get("/")
        self.site = get_current_site(self.request)

        self.user = User.objects.create(username="test", site=self.site)

    def test_authenticate_callback_bad_method(self):
        response = self.client.get(reverse("sites_microsoft_auth:auth-callback"))

        self.assertEqual(405, response.status_code)

    def test_authenticate_callback_no_params(self):
        response = self.client.post(reverse("sites_microsoft_auth:auth-callback"))
        message = json.loads(response.context["message"])["microsoft_auth"]

        self.assertEqual(400, response.status_code)
        self.assertEqual("bad_state", message["error"])
        self.assertEqual(
            AuthenticateCallbackView.messages["bad_state"],
            message["error_description"],
        )

    def test_authenticate_callback_bad_state_format(self):
        response = self.client.post(
            reverse("sites_microsoft_auth:auth-callback"), {"state": "test"}
        )
        message = json.loads(response.context["message"])["microsoft_auth"]

        self.assertEqual(400, response.status_code)
        self.assertEqual("bad_state", message["error"])
        self.assertEqual(
            AuthenticateCallbackView.messages["bad_state"],
            message["error_description"],
        )

    def test_authenticate_callback_bad_state_length(self):
        response = self.client.post(
            reverse("sites_microsoft_auth:auth-callback"), {"state": "001464"}
        )
        message = json.loads(response.context["message"])["microsoft_auth"]

        self.assertEqual(400, response.status_code)
        self.assertEqual("bad_state", message["error"])
        self.assertEqual(
            AuthenticateCallbackView.messages["bad_state"],
            message["error_description"],
        )

    def test_authenticate_callback_bad_state(self):
        response = self.client.post(
            reverse("sites_microsoft_auth:auth-callback"), {"state": STATE[:-1]}
        )
        message = json.loads(response.context["message"])["microsoft_auth"]

        self.assertEqual(400, response.status_code)
        self.assertEqual("bad_state", message["error"])
        self.assertEqual(
            AuthenticateCallbackView.messages["bad_state"],
            message["error_description"],
        )

    def test_authenticate_callback_bad_state_expired(self):
        response = self.client.post(
            reverse("sites_microsoft_auth:auth-callback"), {"state": EXPIRED_STATE}
        )
        message = json.loads(response.context["message"])["microsoft_auth"]

        self.assertEqual(400, response.status_code)
        self.assertEqual("bad_state", message["error"])
        self.assertEqual(
            AuthenticateCallbackView.messages["bad_state"],
            message["error_description"],
        )

    def test_authenticate_callback_missing_code(self):

        response = self.client.post(
            reverse("sites_microsoft_auth:auth-callback"), {"state": STATE}
        )
        message = json.loads(response.context["message"])["microsoft_auth"]

        self.assertEqual(400, response.status_code)
        self.assertEqual("missing_code", message["error"])
        self.assertEqual(
            AuthenticateCallbackView.messages["missing_code"],
            message["error_description"],
        )

    def test_authenticate_callback_error(self):
        response = self.client.post(
            reverse("sites_microsoft_auth:auth-callback"),
            {
                "state": STATE,
                "error": TEST_ERROR,
                "error_description": TEST_ERROR_DESCRIPTION,
            },
        )
        message = json.loads(response.context["message"])["microsoft_auth"]

        self.assertEqual(400, response.status_code)
        self.assertEqual(TEST_ERROR, message["error"])
        self.assertEqual(TEST_ERROR_DESCRIPTION, message["error_description"])

    @patch("sites_microsoft_auth.views.authenticate")
    def test_authenticate_callback_fail_auth(self, mock_auth):
        mock_auth.return_value = None

        response = self.client.post(
            reverse("sites_microsoft_auth:auth-callback"),
            {"state": STATE, "code": "test_code"},
        )
        message = json.loads(response.context["message"])["microsoft_auth"]

        self.assertEqual(400, response.status_code)
        self.assertEqual("login_failed", message["error"])
        self.assertEqual(
            AuthenticateCallbackView.messages["login_failed"],
            message["error_description"],
        )

    @patch("sites_microsoft_auth.views.authenticate")
    @patch("sites_microsoft_auth.views.login")
    def test_authenticate_callback_success(self, mock_login, mock_auth):
        mock_auth.return_value = self.user

        response = self.client.post(
            reverse("sites_microsoft_auth:auth-callback"),
            {"state": STATE, "code": "test_code"},
        )
        message = json.loads(response.context["message"])["microsoft_auth"]

        self.assertEqual(200, response.status_code)
        self.assertEqual({}, message)
        mock_login.assert_called_with(response.wsgi_request, self.user)

    @patch("sites_microsoft_auth.views.authenticate")
    @patch("sites_microsoft_auth.views.login")
    @patch("sites_microsoft_auth.views.get_hook")
    def test_callback_hook(self, mock_get_hook, mock_login, mock_auth):
        def callback(request, context):
            return context

        mock_auth.return_value = self.user

        mock_hook = Mock(side_effect=callback)
        mock_get_hook.return_value = mock_hook

        response = self.client.post(
            reverse("sites_microsoft_auth:auth-callback"),
            {"state": STATE, "code": "test_code"},
        )

        expected_context = {}
        expected_context["base_url"] = response.context["base_url"]
        expected_context["message"] = response.context["message"]

        self.assertIsInstance(expected_context, dict)
        mock_hook.assert_called_with(response.wsgi_request, expected_context)



from django.test import TestCase as BaseTestCase
from django.test import TransactionTestCase as BaseTransactionTestCase


class TestServerSiteMixin:
    def setUp(self):
        from django.contrib.sites.models import Site

        self.site = Site.objects.first()
        self.site.domain = "testserver"
        self.site.name = "testserver"
        self.site.save()


class TestCase(TestServerSiteMixin, BaseTestCase):
    pass


class TransactionTestCase(TestServerSiteMixin, BaseTransactionTestCase):
    pass



from django.conf.urls import include, url
from django.contrib import admin

urlpatterns = [
    url(r"^accounts/", include("django.contrib.auth.urls")),
    url(r"^admin/", admin.site.urls),
    url(
        r"^microsoft/",
        include("sites_microsoft_auth.urls", namespace="sites_microsoft_auth"),
    ),
]



from .old_conf import config

app_name = "sites_microsoft_auth"

urlpatterns = []

if config.MICROSOFT_AUTH_LOGIN_ENABLED:  # pragma: no branch
    from django.conf.urls import url
    from . import views

    urlpatterns = [
        url(
            r"^auth-callback/$",
            views.AuthenticateCallbackView.as_view(),
            name="auth-callback",
        )
    ]



import json
import logging
import re

from django.contrib.auth import authenticate, login
from django.contrib.sites.models import Site
from django.core.signing import BadSignature, SignatureExpired, TimestampSigner
from django.middleware.csrf import CSRF_TOKEN_LENGTH
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from .utils import get_hook, get_scheme

logger = logging.getLogger("django")


class AuthenticateCallbackView(View):
    """ Authentication callback for Microsoft to call as part of OAuth2
            implicit grant flow

        For more details:
        https://developer.microsoft.com/en-us/graph/docs/get-started/rest
    """

    messages = {
        "bad_state": _(
            "An invalid state variable was provided. "
            "Please refresh the page and try again later."
        ),
        "missing_code": _(
            "No authentication code was provided from "
            "Microsoft. Please try again."
        ),
        "login_failed": _(
            "Failed to authenticate you for an unknown reason. "
            "Please try again later."
        ),
    }

    # manually mark methods csrf_exempt to handle CSRF processing ourselves
    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        domain = Site.objects.get_current(self.request).domain

        scheme = get_scheme(self.request)

        self.context = {
            "base_url": "{0}://{1}/".format(scheme, domain),
            "message": {},
        }

        # validates state using Django CSRF system
        self._check_csrf(kwargs.get("state"))

        # validates response from Microsoft
        self._check_microsoft_response(
            kwargs.get("error"), kwargs.get("error_description")
        )

        # validates the code param and logs user in
        self._authenticate(kwargs.get("code"))

        # populates error_description if it does not exist yet
        if (
            "error" in self.context["message"]
            and "error_description" not in self.context["message"]
        ):
            self.context["message"]["error_description"] = self.messages[
                self.context["message"]["error"]
            ]

        function = get_hook("MICROSOFT_AUTH_CALLBACK_HOOK")
        if function is not None:
            self.context = function(self.request, self.context)

        self.context["message"] = mark_safe(  # nosec
            json.dumps({"microsoft_auth": self.context["message"]})
        )
        return self.context

    def _check_csrf(self, state):
        signer = TimestampSigner()

        if state is None:
            state = ""

        try:
            state = signer.unsign(state, max_age=300)
        except BadSignature:  # pragma: no branch
            logger.debug("state has been tempered with")
            state = ""
        except SignatureExpired:  # pragma: no cover
            logger.debug("state has expired")
            state = ""

        checks = (
            re.search("[a-zA-Z0-9]", state),
            len(state) == CSRF_TOKEN_LENGTH,
        )

        # validate state parameter
        if not all(checks):
            logger.debug("State validation failed:")
            logger.debug("state: {}".format(state))
            logger.debug("checks: {}".format(checks))
            self.context["message"] = {"error": "bad_state"}

    def _check_microsoft_response(self, error, error_description):
        if "error" not in self.context["message"]:
            if error is not None:
                self.context["message"] = {
                    "error": error,
                    "error_description": error_description,
                }

    def _authenticate(self, code):
        if "error" not in self.context["message"]:
            if code is None:
                self.context["message"] = {"error": "missing_code"}
            else:
                # authenticate user using Microsoft code
                user = authenticate(self.request, code=code)
                if user is None:
                    # this should not fail at this point except for network
                    # error while retrieving profile or database error
                    # adding new user
                    self.context["message"] = {"error": "login_failed"}
                else:
                    login(self.request, user)

    def post(self, request):
        """ main callback for Microsoft to call

            validates Microsoft response, attempts to authenticate user and
            returns simple HTML page with Javascript that will post a message
            to parent window with details of result """

        context = self.get_context_data(**request.POST.dict())

        status_code = 200
        if "error" in context["message"]:
            status_code = 400

        return render(
            request,
            "microsoft/auth_callback.html",
            context,
            status=status_code,
        )



from unittest.mock import Mock, patch

from django.contrib.auth import authenticate, get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.test import RequestFactory, override_settings

from sites_microsoft_auth.conf import get_conf
from sites_microsoft_auth.models import MicrosoftAccount

from .. import TestCase

CODE = "test_code"
TOKEN = {"access_token": "test_token", "scope": ["test"]}
EMAIL = "some.email@example.com"
EMAIL2 = "some.email2@example.com"
FIRST = "Test"
LAST = "User"
MISSING_ID = "some_missing_id"


@override_settings(
    AUTHENTICATION_BACKENDS=[
        "sites_microsoft_auth.backends.MicrosoftAuthenticationBackend",
        "django.contrib.auth.backends.ModelBackend",
    ]
)
class MicrosoftBackendsTests(TestCase):
    def setUp(self):
        super().setUp()

        User = get_user_model()

        self.factory = RequestFactory()
        self.request = self.factory.get("/")
        self.site = get_current_site(self.request)

        self.linked_account = MicrosoftAccount.objects.create(
            microsoft_id="test_id", site=self.site
        )
        self.linked_account.user = User.objects.create(
            username="user1", email=EMAIL2, site=self.site
        )
        self.linked_account.save()

        self.unlinked_account = MicrosoftAccount.objects.create(
            microsoft_id="missing_id", site=self.site
        )

        self.unlinked_user = User.objects.create(
            username="user2", email="test@example.com", site=self.site
        )

    def test_authenticate_no_code(self):
        user = authenticate(self.request)

        self.assertIs(user, None)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_invalid_token(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = {}

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertTrue(mock_auth.fetch_token.called)
        self.assertIs(user, None)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_invalid_scopes(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = False

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertTrue(mock_auth.fetch_token.called)
        self.assertTrue(mock_auth.valid_scopes.called)
        self.assertIs(user, None)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_invalid_profile(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = None

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertTrue(mock_auth.fetch_token.called)
        self.assertTrue(mock_auth.valid_scopes.called)
        self.assertIs(user, None)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_errored_profile(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = None

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertTrue(mock_auth.fetch_token.called)
        self.assertTrue(mock_auth.valid_scopes.called)
        self.assertIs(user, None)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_user(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": self.linked_account.microsoft_id
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.linked_account.user.id)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_user_missing_user(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": self.unlinked_account.microsoft_id,
            "email": EMAIL,
            "name": "{} {}".format(FIRST, LAST),
            "preferred_username": EMAIL,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)
        self.unlinked_account.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.unlinked_account.user.id)
        self.assertEqual(EMAIL, self.unlinked_account.user.email)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_user_no_user_with_name(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": self.unlinked_account.microsoft_id,
            "email": EMAIL,
            "name": "{} {}".format(FIRST, LAST),
            "preferred_username": EMAIL,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)
        self.unlinked_account.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.unlinked_account.user.id)
        self.assertEqual(EMAIL, self.unlinked_account.user.email)
        self.assertEqual(FIRST, self.unlinked_account.user.first_name)
        self.assertEqual(LAST, self.unlinked_account.user.last_name)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_user_unlinked_user(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": self.unlinked_account.microsoft_id,
            "email": self.unlinked_user.email,
            "name": "{} {}".format(FIRST, LAST),
            "preferred_username": EMAIL,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)
        self.unlinked_account.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.unlinked_account.user.id)
        self.assertEqual(self.unlinked_user.id, self.unlinked_account.user.id)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_user_unlinked_user_no_autofill(
        self, mock_client
    ):
        expected_first_name = "Test"
        expected_last_name = "User"

        self.unlinked_user.first_name = expected_first_name
        self.unlinked_user.last_name = expected_last_name
        self.unlinked_user.save()

        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": self.unlinked_account.microsoft_id,
            "email": self.unlinked_user.email,
            "name": "{} {}".format(FIRST, LAST),
            "preferred_username": EMAIL,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)
        self.unlinked_account.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.unlinked_account.user.id)
        self.assertEqual(self.unlinked_user.id, self.unlinked_account.user.id)
        self.assertEqual(expected_first_name, self.unlinked_user.first_name)
        self.assertEqual(expected_last_name, self.unlinked_user.last_name)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_user_unlinked_user_no_name(
        self, mock_client
    ):
        self.unlinked_user.save()

        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": self.unlinked_account.microsoft_id,
            "email": self.unlinked_user.email,
            "preferred_username": EMAIL,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)
        self.unlinked_account.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.unlinked_account.user.id)
        self.assertEqual(self.unlinked_user.id, self.unlinked_account.user.id)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_user_missing_name(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": self.unlinked_account.microsoft_id,
            "email": self.unlinked_user.email,
            "name": "{} {}".format(FIRST, LAST),
            "preferred_username": EMAIL,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)
        self.unlinked_user.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.unlinked_user.id)
        self.assertEqual(FIRST, self.unlinked_user.first_name)
        self.assertEqual(LAST, self.unlinked_user.last_name)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_no_autocreate(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {"sub": MISSING_ID}

        mock_client.return_value = mock_auth

        config = get_conf(self.request)
        config.auto_create = False
        config.save()

        user = authenticate(self.request, code=CODE)

        self.assertIs(user, None)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_autocreate(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": MISSING_ID,
            "email": EMAIL,
            "name": "{} {}".format(FIRST, LAST),
            "preferred_username": EMAIL,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertIsNot(user, None)
        self.assertEqual(EMAIL, user.email)
        self.assertEqual(MISSING_ID, user.microsoft_account.microsoft_id)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_autocreate_no_email(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": MISSING_ID,
            "name": "{} {}".format(FIRST, LAST),
            "preferred_username": EMAIL,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertIsNot(user, None)
        self.assertEqual(EMAIL, user.username)
        self.assertEqual(MISSING_ID, user.microsoft_account.microsoft_id)
        self.assertEqual(user.email, '')

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_linked_default(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": MISSING_ID,
            "email": EMAIL2,
            "name": "{} {}".format(FIRST, LAST),
            "preferred_username": EMAIL,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertIsNone(user)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_linked_replace(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": MISSING_ID,
            "email": EMAIL2,
            "name": "{} {}".format(FIRST, LAST),
            "preferred_username": EMAIL,
        }

        config = get_conf(self.request)
        config.auto_replace_accounts = True
        config.save()

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)
        self.linked_account.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(EMAIL2, user.email)
        self.assertEqual(MISSING_ID, user.microsoft_account.microsoft_id)
        self.assertIsNone(self.linked_account.user)

    @override_settings(
        MICROSOFT_AUTH_AUTHENTICATE_HOOK="tests.test_backends.test_microsoft.hook_callback"  # noqa
    )
    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    @patch("sites_microsoft_auth.backends.get_hook")
    def test_authenticate_hook(self, mock_get_hook, mock_client):
        mock_hook = Mock()
        mock_get_hook.return_value = mock_hook

        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_claims.return_value = {
            "sub": self.unlinked_account.microsoft_id,
            "email": self.unlinked_user.email,
            "name": "{} {}".format(FIRST, LAST),
            "preferred_username": EMAIL,
        }
        mock_auth.token = TOKEN

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)
        self.unlinked_account.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.unlinked_account.user.id)
        self.assertEqual(self.unlinked_user.id, self.unlinked_account.user.id)

        mock_hook.assert_called_with(user, TOKEN)



from unittest.mock import Mock, patch

from django.test import RequestFactory, override_settings

from sites_microsoft_auth.old_conf import config
from sites_microsoft_auth.utils import get_hook, get_scheme

from . import TestCase


class UtilsTests(TestCase):
    def setUp(self):
        super().setUp()

        self.factory = RequestFactory()
        self.request = self.factory.get("/")

    def test_default(self):
        self.assertEqual(get_scheme(self.request), "https")

    def test_config_param(self):
        self.assertEqual(get_scheme(self.request, config), "https")

    @override_settings(DEBUG=True)
    def test_debug(self):
        self.assertEqual(get_scheme(self.request), "http")

    @override_settings(DEBUG=True)
    def test_debug_no_request(self):
        self.assertEqual(get_scheme(None), "https")

    @override_settings(DEBUG=True)
    def test_debug_forwarded_proto(self):
        self.request.META["HTTP_X_FORWARDED_PROTO"] = "https"

        self.assertEqual(get_scheme(self.request), "https")

    def test_get_hook_invalid(self):
        self.assertIsNone(get_hook("NOT_A_REAL_HOOK"))

    def test_get_hook_valid_default(self):
        self.assertIsNone(get_hook("MICROSOFT_AUTH_AUTHENTICATE_HOOK"))

    @override_settings(
        MICROSOFT_AUTH_AUTHENTICATE_HOOK="tests.test_utils.hook_callback"
    )
    @patch("sites_microsoft_auth.utils.importlib")
    def test_get_hook_valid_not_empty(self, mock_import):
        mock_module = Mock()
        mock_module.hook_callback = Mock()
        mock_import.import_module.return_value = mock_module

        function = get_hook("MICROSOFT_AUTH_AUTHENTICATE_HOOK")

        self.assertEqual(function, mock_module.hook_callback)



# Generated by Django 2.2.5 on 2019-09-23 15:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('sites_microsoft_auth', '0004_auto_20190923_1512'),
    ]

    operations = [
        migrations.AlterField(
            model_name='microsoftaccount',
            name='site',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='sites.Site'),
        ),
        migrations.AlterField(
            model_name='xboxliveaccount',
            name='site',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='sites.Site'),
        ),
    ]



from django.contrib.sites.models import Site
from django.db.models.signals import post_save
from django.dispatch import receiver

from sites_microsoft_auth.models import SiteConfiguration


@receiver(post_save, sender=Site)
def create_site_config(sender, instance, **kwargs):
    SiteConfiguration.objects.get_or_create(site=instance)



from django.apps import apps
from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin

from .old_conf import LOGIN_TYPE_MA, LOGIN_TYPE_XBL, config
from .models import MicrosoftAccount, XboxLiveAccount, SiteConfiguration

__all__ = [
    "MicrosoftAccountAdmin",
    "MicrosoftAccountInlineAdmin",
    "UserAdmin",
    "XboxLiveAccountAdmin",
    "XboxLiveAccountInlineAdmin",
]

User = get_user_model()

# override admin site template
admin.site.login_template = "microsoft/admin_login.html"

# djangoql support
extra_base = []
if apps.is_installed("djangoql"):  # pragma: no branch
    from djangoql.admin import DjangoQLSearchMixin

    extra_base = [DjangoQLSearchMixin]

base_admin = extra_base + [admin.ModelAdmin]
base_user_admin = extra_base + [BaseUserAdmin]

# unregister User mode if it is already registered
if admin.site.is_registered(User):  # pragma: no branch
    admin.site.unregister(User)


class MicrosoftAccountAdmin(*base_admin):
    readonly_fields = ("microsoft_id", "site",)


class MicrosoftAccountInlineAdmin(admin.StackedInline):
    model = MicrosoftAccount
    readonly_fields = ("microsoft_id", "site",)


class XboxLiveAccountAdmin(*base_admin):
    readonly_fields = ("xbox_id", "gamertag", "site")


class XboxLiveAccountInlineAdmin(admin.StackedInline):
    model = XboxLiveAccount
    readonly_fields = ("xbox_id", "gamertag", "site")


def _register_admins():
    _do_both = config.MICROSOFT_AUTH_REGISTER_INACTIVE_ADMIN
    _login_type = config.MICROSOFT_AUTH_LOGIN_TYPE

    if admin.site.is_registered(MicrosoftAccount):
        admin.site.unregister(MicrosoftAccount)

    if admin.site.is_registered(XboxLiveAccount):
        admin.site.unregister(XboxLiveAccount)

    if admin.site.is_registered(SiteConfiguration):
        admin.site.unregister(SiteConfiguration)

    if _do_both or _login_type == LOGIN_TYPE_MA:
        admin.site.register(MicrosoftAccount, MicrosoftAccountAdmin)
    if _do_both or _login_type == LOGIN_TYPE_XBL:
        admin.site.register(XboxLiveAccount, XboxLiveAccountAdmin)
    admin.site.register(SiteConfiguration)


def _get_inlines():
    _do_both = config.MICROSOFT_AUTH_REGISTER_INACTIVE_ADMIN
    _login_type = config.MICROSOFT_AUTH_LOGIN_TYPE
    inlines = []

    if _do_both or _login_type == LOGIN_TYPE_MA:
        inlines.append(MicrosoftAccountInlineAdmin)
    if _do_both or _login_type == LOGIN_TYPE_XBL:
        inlines.append(XboxLiveAccountInlineAdmin)

    return inlines


@admin.register(User)
class UserAdmin(*base_user_admin):
    @property
    def inlines(self):
        """ Adds MicrosoftAccount and/or XboxLiveAccount foreign keys to
        User model """

        return _get_inlines()


_register_admins()



from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.contrib.sites.models import Site
from django.db import models
from django.utils.translation import ugettext_lazy as _


class UnicodeSpaceUsernameValidator(UnicodeUsernameValidator):
    """ validator to allow spaces in username """

    regex = r"^[\w\.@+\- ]+$"


class SitesUser(AbstractUser):
    username = models.CharField(
        _('username'),
        max_length=150,
        help_text=_('Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.'),
        validators=[UnicodeSpaceUsernameValidator],
        unique=True,
    )
    site = models.ForeignKey(Site, on_delete=models.CASCADE)

    class Meta:
        abstract = True


class MicrosoftAccount(models.Model):
    microsoft_id = models.CharField(_("microsoft account id"), max_length=64)
    site = models.ForeignKey(Site, on_delete=models.CASCADE)
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        null=True,
        related_name="microsoft_account",
    )

    def __str__(self):
        return self.microsoft_id


class XboxLiveAccount(models.Model):
    xbox_id = models.CharField(_("xbox user id"), max_length=32, unique=True)
    site = models.ForeignKey(Site, on_delete=models.CASCADE)
    gamertag = models.CharField(_("xbox live gamertag"), max_length=16)
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        null=True,
        related_name="xbox_live_account",
    )

    def __str__(self):
        return self.gamertag


class SiteConfiguration(models.Model):
    LOGIN_TYPE_MA = "ma"
    LOGIN_TYPE_XBL = "xbl"
    LOGIN_TYPE_CHOICES = (
        (LOGIN_TYPE_MA, 'Microsoft Account'),
        (LOGIN_TYPE_XBL, 'Xbox Live Account'),
    )

    site = models.OneToOneField(to=Site, on_delete=models.CASCADE)
    login_enabled = models.BooleanField(default=False)
    login_type = models.CharField(max_length=2, choices=LOGIN_TYPE_CHOICES)

    # common means any microsoft account otherwise Microsoft Office 365 Tenant ID
    tenant_id = models.CharField(max_length=80, default='common')
    # Microsoft OAuth Client ID, see https://apps.dev.microsoft.com/ for more.
    client_id = models.CharField(max_length=80, default='', blank=True)
    # Microsoft OAuth Client Secret, see https://apps.dev.microsoft.com/ for more.
    client_secret = models.CharField(max_length=80, default='', blank=True)
    # Extra OAuth scopes for authentication. Required scopes are always provided ('openid email' for Microsoft Auth
    # and 'XboxLive.signin XboxLive.offline_access' for Xbox). Scopes are space delimited.
    extra_scopes = models.CharField(max_length=80, default='', blank=True)
    # Autocreate user that attempt to login if they do not already exist?
    auto_create = models.BooleanField(default=True)
    # Automatically register admin class for auth type that is not active (Xbox when Microsoft Auth is enabled
    # and Microsoft Auth when Xbox is enabled). Requires restart of app for setting to take effect.
    register_inactive_admin = models.BooleanField(default=False)
    # Automatically sync the username from the Xbox Live Gamertag?
    xbl_sync_username = models.BooleanField(default=False)
    # Automatically replace an existing Microsoft Account paired to a user when authenticating.
    auto_replace_accounts = models.BooleanField(default=False)
    # Callable hook to call after authenticating a user on the `sites_microsoft_auth.backends.MicrosoftAuthenticationBackend`.
    # If the login type is Microsoft Auth, the parameters will be
    # `(User: user, oauthlib.oauth2.rfc6749.tokens.OAuth2Token: token)`
    # If the login type is Xbox Live, the parameters will be `(User:user, dict: token)`
    # where token is the Xbox Token, see `sites_microsoft_auth.client.MicrosoftClient.fetch_xbox_token` for format
    authenticate_hook = models.CharField(max_length=100, default='', blank=True)
    # """Callable hook to call right before completing the `auth_callback` view. Really useful for adding custom data
    # to message or chaining the expected base URL that gets passed back up to the window that initiated the original
    # Authorize request. The parameters that will be passed will be `(HttpRequest: request, dict: context)`.
    # The expected return value is the updated context dictionary. You should NOT remove the data that is currently
    # there. `base_url` is the expected root URL of the window that initiated the authorize request `message` is
    # a dictionary that will be serialized as a JSON string and passoed back to the initiating window.
    callback_hook = models.CharField(max_length=100, default='', blank=True)

    def __getattr__(self, name):
        """
        Allow previous MICROSOFT_AUTH_ format for accessing instance variables.
        :param name:
        :return:
        """
        real_name = name.replace('MICROSOFT_AUTH_', '').lower()
        return super().__getattribute__(real_name)



#!/usr/bin/env python
import os
import sys


def manage():
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.site.settings")
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc

    if "runserver" in sys.argv:
        os.environ["DEBUG_SKIP_THREAD"] = "1"
    os.environ["LOAD_DEV_SETTINGS"] = "1"
    execute_from_command_line(sys.argv)


if __name__ == "__main__":
    manage()



""" isort:skip_file """

from unittest.mock import Mock, patch

import pytest
from django.test import RequestFactory, override_settings

from sites_microsoft_auth.conf import get_conf
from sites_microsoft_auth.old_conf import LOGIN_TYPE_XBL
from sites_microsoft_auth.context_processors import microsoft

from . import TestCase

URL = "https://example.com"


@override_settings(DEBUG=True)
class ContextProcessorsTests(TestCase):
    @pytest.fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def setUp(self):
        super().setUp()

        self.factory = RequestFactory()

    @patch("sites_microsoft_auth.context_processors.MicrosoftClient")
    def test_microsoft_login_enabled(self, mock_client):
        request = self.factory.get("/")
        config = get_conf(request)
        config.login_enabled = True
        config.save()
        context = microsoft(request)

        self.assertTrue(context.get("microsoft_login_enabled"))

    @patch("sites_microsoft_auth.context_processors.MicrosoftClient")
    def test_microsoft_login_enabled_disabled(self, mock_client):
        request = self.factory.get("/")
        context = microsoft(request)

        self.assertFalse(context.get("microsoft_login_enabled"))

    @patch("sites_microsoft_auth.context_processors.MicrosoftClient")
    @patch("sites_microsoft_auth.context_processors.mark_safe")
    def test_microsoft_authorization_url(self, mock_safe, mock_client):
        mock_client_i = Mock()
        mock_client_i.authorization_url.return_value = [URL]
        mock_client.return_value = mock_client_i
        mock_safe.side_effect = lambda value: value

        request = self.factory.get("/")
        context = microsoft(request)

        self.assertEqual(URL, context.get("microsoft_authorization_url"))

    @patch("sites_microsoft_auth.context_processors.MicrosoftClient")
    def test_microsoft_login_type_text(self, mock_client):

        request = self.factory.get("/")
        context = microsoft(request)

        self.assertEqual("Microsoft", context.get("microsoft_login_type_text"))

    @patch("sites_microsoft_auth.context_processors.MicrosoftClient")
    def test_microsoft_login_type_text_xbl(self, mock_client):

        request = self.factory.get("/")

        config = get_conf(request)
        config.login_type = LOGIN_TYPE_XBL
        config.save()
        context = microsoft(request)

        self.assertEqual("Xbox Live", context.get("microsoft_login_type_text"))



from unittest.mock import Mock, patch

from django.contrib.auth import authenticate, get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.test import RequestFactory, override_settings

from sites_microsoft_auth.conf import get_conf
from sites_microsoft_auth.old_conf import LOGIN_TYPE_XBL
from sites_microsoft_auth.models import XboxLiveAccount

from .. import TestCase

CODE = "test_code"
TOKEN = {"access_token": "test_token", "scope": ["test"]}
XBOX_TOKEN = {"Token": "test_token"}
LAST = "User"
MISSING_ID = "some_missing_id"
GAMERTAG = "Some Gamertag"


@override_settings(
    AUTHENTICATION_BACKENDS=[
        "sites_microsoft_auth.backends.MicrosoftAuthenticationBackend",
        "django.contrib.auth.backends.ModelBackend",
    ],
)
class XboxLiveBackendsTests(TestCase):
    def setUp(self):
        super().setUp()

        User = get_user_model()

        self.factory = RequestFactory()
        self.request = self.factory.get("/")
        self.site = get_current_site(self.request)

        config = get_conf(self.request)
        config.login_type = LOGIN_TYPE_XBL
        config.save()

        self.linked_account = XboxLiveAccount.objects.create(
            xbox_id="test_id", gamertag="test_gamertag", site=self.site
        )
        self.linked_account.user = User.objects.create(username="user1", site=self.site)
        self.linked_account.save()

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_bad_xbox_token(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.fetch_xbox_token.return_value = {}
        mock_auth.valid_scopes.return_value = True

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertFalse(mock_auth.get_xbox_profile.called)
        self.assertIs(user, None)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_user(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.fetch_xbox_token.return_value = XBOX_TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_xbox_profile.return_value = {
            "xid": self.linked_account.xbox_id,
            "gtg": self.linked_account.gamertag,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.linked_account.user.id)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_existing_user_new_gamertag(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.fetch_xbox_token.return_value = XBOX_TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_xbox_profile.return_value = {
            "xid": self.linked_account.xbox_id,
            "gtg": GAMERTAG,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)
        self.linked_account.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.linked_account.user.id)
        self.assertEqual(GAMERTAG, self.linked_account.gamertag)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_no_autocreate(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.fetch_xbox_token.return_value = XBOX_TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_xbox_profile.return_value = {
            "xid": MISSING_ID,
            "gtg": GAMERTAG,
        }

        conf = get_conf(self.request)
        conf.auto_create = False
        conf.save()

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertIs(user, None)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_autocreate(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.fetch_xbox_token.return_value = XBOX_TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_xbox_profile.return_value = {
            "xid": MISSING_ID,
            "gtg": GAMERTAG,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        self.assertIsNot(user, None)
        self.assertEqual(GAMERTAG, user.username)
        self.assertEqual(MISSING_ID, user.xbox_live_account.xbox_id)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_no_sync_username(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.fetch_xbox_token.return_value = XBOX_TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_xbox_profile.return_value = {
            "xid": self.linked_account.xbox_id,
            "gtg": GAMERTAG,
        }

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)
        self.linked_account.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.linked_account.user.id)
        self.assertNotEqual(GAMERTAG, self.linked_account.user.username)

    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    def test_authenticate_sync_username(self, mock_client):
        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.fetch_xbox_token.return_value = XBOX_TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_xbox_profile.return_value = {
            "xid": self.linked_account.xbox_id,
            "gtg": GAMERTAG,
        }

        mock_client.return_value = mock_auth

        conf = get_conf(self.request)
        conf.xbl_sync_username = True
        conf.save()

        user = authenticate(self.request, code=CODE)
        self.linked_account.refresh_from_db()
        self.linked_account.user.refresh_from_db()

        self.assertIsNot(user, None)
        self.assertEqual(user.id, self.linked_account.user.id)
        self.assertEqual(GAMERTAG, self.linked_account.user.username)

    @override_settings(
        MICROSOFT_AUTH_AUTHENTICATE_HOOK="tests.test_backends.test_xbox.hook_callback"  # noqa
    )
    @patch("sites_microsoft_auth.backends.MicrosoftClient")
    @patch("sites_microsoft_auth.backends.get_hook")
    def test_authenticate_hook(self, mock_get_hook, mock_client):
        mock_hook = Mock()
        mock_get_hook.return_value = mock_hook

        mock_auth = Mock()
        mock_auth.fetch_token.return_value = TOKEN
        mock_auth.fetch_xbox_token.return_value = XBOX_TOKEN
        mock_auth.valid_scopes.return_value = True
        mock_auth.get_xbox_profile.return_value = {
            "xid": MISSING_ID,
            "gtg": GAMERTAG,
        }
        mock_auth.xbox_token = XBOX_TOKEN

        mock_client.return_value = mock_auth

        user = authenticate(self.request, code=CODE)

        mock_hook.assert_called_with(user, XBOX_TOKEN)



import logging

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.contrib.sites.shortcuts import get_current_site

from sites_microsoft_auth import conf
from .client import MicrosoftClient
from .conf import LOGIN_TYPE_XBL
from .models import MicrosoftAccount, XboxLiveAccount
from .utils import get_hook

logger = logging.getLogger("django")
User = get_user_model()


class MicrosoftAuthenticationBackend(ModelBackend):
    """ Authentication backend to authenticate a user against their Microsoft
        Uses Microsoft's Graph OAuth and XBL servers to authentiate. """

    config = None
    microsoft = None
    site = None

    def authenticate(self, request, code=None):
        """
            Authenticates the user against the Django backend
                using a Microsoft auth code from
            https://login.microsoftonline.com/common/oauth2/v2.0/authorize or
            https://login.live.com/oauth20_authorize.srf

            For more details:
            https://developer.microsoft.com/en-us/graph/docs/get-started/rest
        """

        self.microsoft = MicrosoftClient(request=request)
        self.config = conf.get_conf(request)
        self.site = get_current_site(request)

        user = None
        if code is not None:
            # fetch OAuth token
            token = self.microsoft.fetch_token(code=code)

            # validate permission scopes
            if "access_token" in token and self.microsoft.valid_scopes(
                token["scope"]
            ):
                user = self._authenticate_user()

        if user is not None:
            self._call_hook(user)

        return user

    def _authenticate_user(self):
        if self.config.MICROSOFT_AUTH_LOGIN_TYPE == LOGIN_TYPE_XBL:
            return self._authenticate_xbox_user()
        else:
            return self._authenticate_microsoft_user()

    def _authenticate_xbox_user(self):
        xbox_token = self.microsoft.fetch_xbox_token()

        if "Token" in xbox_token:
            response = self.microsoft.get_xbox_profile()
            return self._get_user_from_xbox(response)
        return None

    def _authenticate_microsoft_user(self):
        claims = self.microsoft.get_claims()

        if claims is not None:
            return self._get_user_from_microsoft(claims)

        return None

    def _get_user_from_xbox(self, data):
        """ Retrieves existing Django user or creates
            a new one from Xbox Live profile data """
        user = None
        xbox_user = self._get_xbox_user(data)

        if xbox_user is not None:
            self._verify_xbox_user(xbox_user)

            user = xbox_user.user

            if (
                self.config.MICROSOFT_AUTH_XBL_SYNC_USERNAME
                and user.username != xbox_user.gamertag
            ):
                user.username = xbox_user.gamertag
                user.save()

        return user

    def _get_xbox_user(self, data):
        xbox_user = None

        try:
            xbox_user = XboxLiveAccount.objects.get(xbox_id=data["xid"], site=self.site)
            # update Gamertag since they can change over time
            if xbox_user.gamertag != data["gtg"]:
                xbox_user.gamertag = data["gtg"]
                xbox_user.save()
        except XboxLiveAccount.DoesNotExist:
            if self.config.MICROSOFT_AUTH_AUTO_CREATE:
                # create new Xbox Live Account
                xbox_user = XboxLiveAccount(
                    xbox_id=data["xid"], gamertag=data["gtg"], site=self.site
                )
                xbox_user.save()

        return xbox_user

    def _verify_xbox_user(self, xbox_user):
        if xbox_user.user is None:
            user = User(username=xbox_user.gamertag, site=self.site)
            user.save()

            xbox_user.user = user
            xbox_user.save()

    def _get_user_from_microsoft(self, data):
        """ Retrieves existing Django user or creates
            a new one from Xbox Live profile data """
        user = None
        microsoft_user = self._get_microsoft_user(data)

        if microsoft_user is not None:
            user = self._verify_microsoft_user(microsoft_user, data)

        return user

    def _get_microsoft_user(self, data):
        microsoft_user = None

        try:
            microsoft_user = MicrosoftAccount.objects.get(
                microsoft_id=data["sub"], site=self.site
            )
        except MicrosoftAccount.DoesNotExist:
            if self.config.MICROSOFT_AUTH_AUTO_CREATE:
                # create new Microsoft Account
                microsoft_user = MicrosoftAccount(microsoft_id=data["sub"], site=self.site)
                microsoft_user.save()

        return microsoft_user

    def _verify_microsoft_user(self, microsoft_user, data):
        user = microsoft_user.user

        if user is None:
            fullname = data.get("name")
            first_name, last_name = "", ""
            if fullname is not None:
                first_name, last_name = data["name"].split(" ", 1)

            try:
                # create new Django user from provided data
                try:
                    user = User.objects.get(
                        username__iexact=data['preferred_username'][:150],
                        site=self.site,
                    )
                except User.DoesNotExist:
                    user = User.objects.get(
                        email__iexact=data.get('email', None),
                        site=self.site,
                    )

                if user.first_name == "" and user.last_name == "":
                    user.first_name = first_name
                    user.last_name = last_name
                    user.save()
            except User.DoesNotExist:
                user = User(
                    username=data["preferred_username"][:150],
                    first_name=first_name,
                    last_name=last_name,
                    email=data.get('email', ''),
                    site=self.site
                )
                user.save()

            existing_account = self._get_existing_microsoft_account(user)
            if existing_account is not None:
                if self.config.MICROSOFT_AUTH_AUTO_REPLACE_ACCOUNTS:
                    existing_account.user = None
                    existing_account.save()
                else:
                    logger.warning(
                        (
                            "User {} already has linked Microsoft "
                            "account and MICROSOFT_AUTH_AUTO_REPLACE_ACCOUNTS "
                            "is False"
                        ).format(user.email)
                    )
                    return None

            microsoft_user.user = user
            microsoft_user.save()

        return user

    def _get_existing_microsoft_account(self, user):
        try:
            return MicrosoftAccount.objects.get(user=user, site=self.site)
        except MicrosoftAccount.DoesNotExist:
            return None

    def _call_hook(self, user):
        function = get_hook("MICROSOFT_AUTH_AUTHENTICATE_HOOK")
        if function is not None:
            if self.config.MICROSOFT_AUTH_LOGIN_TYPE == LOGIN_TYPE_XBL:
                function(user, self.microsoft.xbox_token)
            else:
                function(user, self.microsoft.token)



import logging

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.signing import TimestampSigner
from django.middleware.csrf import get_token
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from .client import MicrosoftClient
from .utils import get_scheme
from .conf import get_conf, LOGIN_TYPE_XBL

logger = logging.getLogger("django")


def microsoft(request):
    """ Adds global template variables for sites_microsoft_auth """
    login_type = None
    config = get_conf(request)
    if config.MICROSOFT_AUTH_LOGIN_TYPE == LOGIN_TYPE_XBL:
        login_type = _("Xbox Live")
    else:
        login_type = _("Microsoft")

    if settings.DEBUG:  # pragma: no branch
        try:
            current_domain = Site.objects.get_current(request).domain
        except Site.DoesNotExist:
            logger.warning(
                "\nWARNING:\nThe domain configured for the sites framework "
                "does not match the domain you are accessing Django with. "
                "Microsoft authentication may not work.\n"
            )
        else:
            do_warning = get_scheme(
                request
            ) == "http" and not current_domain.startswith("localhost")
            if do_warning:  # pragma: no branch
                logger.warning(
                    "\nWARNING:\nYou are not using HTTPS. Microsoft "
                    "authentication only works over HTTPS unless the hostname "
                    "for your `redirect_uri` is `localhost`\n"
                )

    # initialize Microsoft client using CSRF token as state variable
    signer = TimestampSigner()
    state = signer.sign(get_token(request))
    microsoft = MicrosoftClient(state=state, request=request)
    auth_url = microsoft.authorization_url()[0]
    return {
        "microsoft_login_enabled": config.MICROSOFT_AUTH_LOGIN_ENABLED,
        "microsoft_authorization_url": mark_safe(auth_url),  # nosec
        "microsoft_login_type_text": login_type,
    }



import json
import logging

import jwt
import requests
from django.contrib.sites.models import Site
from django.contrib.sites.shortcuts import get_current_site
from django.core.cache import cache
from django.urls import reverse
from jwt.algorithms import RSAAlgorithm
from requests_oauthlib import OAuth2Session

from .conf import (
    CACHE_KEY_JWKS,
    CACHE_KEY_OPENID,
    CACHE_TIMEOUT,
    LOGIN_TYPE_XBL,
)
from .utils import get_scheme

logger = logging.getLogger("django")


class MicrosoftClient(OAuth2Session):
    """ Simple Microsoft OAuth2 Client to authenticate them

        Extended from Requests-OAuthlib's OAuth2Session class which
            does most of the heavy lifting

        https://requests-oauthlib.readthedocs.io/en/latest/

        Microsoft OAuth documentation can be found at
        https://developer.microsoft.com/en-us/graph/docs/get-started/rest
    """

    _config_url = "https://login.microsoftonline.com/{tenant}/v2.0/.well-known/openid-configuration"  # noqa

    _xbox_authorization_url = "https://login.live.com/oauth20_authorize.srf"
    _xbox_token_url = (
        "https://user.auth.xboxlive.com/user/authenticate"  # nosec
    )
    _profile_url = "https://xsts.auth.xboxlive.com/xsts/authorize"

    xbox_token = {}

    config = None

    # required OAuth scopes
    SCOPE_XBL = ["XboxLive.signin", "XboxLive.offline_access"]
    SCOPE_MICROSOFT = ["openid", "email", "profile"]

    def __init__(self, state=None, request=None, *args, **kwargs):
        from .conf import get_conf

        self.config = get_conf(request)

        extra_scopes = self.config.MICROSOFT_AUTH_EXTRA_SCOPES

        try:
            current_site = Site.objects.get_current(request)
        except Site.DoesNotExist:
            current_site = Site.objects.first()

        self.site = current_site

        domain = current_site.domain
        path = reverse("sites_microsoft_auth:auth-callback")
        scope = " ".join(self.SCOPE_MICROSOFT)

        if self.config.MICROSOFT_AUTH_LOGIN_TYPE == LOGIN_TYPE_XBL:
            scope = " ".join(self.SCOPE_XBL)

        scope = "{} {}".format(scope, extra_scopes).strip()

        scheme = get_scheme(request, self.config)

        super().__init__(
            self.config.MICROSOFT_AUTH_CLIENT_ID,
            scope=scope,
            state=state,
            redirect_uri="{0}://{1}{2}".format(scheme, domain, path),
            *args,
            **kwargs
        )

    @property
    def openid_config(self):
        config = cache.get(CACHE_KEY_OPENID.format(site=self.site))

        if config is None:
            config_url = self._config_url.format(
                tenant=self.config.MICROSOFT_AUTH_TENANT_ID
            )
            response = self.get(config_url)

            if response.ok:
                config = response.json()
                cache.set(CACHE_KEY_OPENID.format(site=self.site), config, CACHE_TIMEOUT)

        return config

    @property
    def jwks(self):
        jwks = cache.get(CACHE_KEY_JWKS.format(site=self.site), [])

        if len(jwks) == 0:
            jwks_uri = self.openid_config["jwks_uri"]
            if jwks_uri is None:
                return []

            response = self.get(jwks_uri)

            if response.ok:
                jwks = response.json()["keys"]
                cache.set(CACHE_KEY_JWKS.format(self.site), jwks, CACHE_TIMEOUT)
        return jwks

    def get_claims(self, allow_refresh=True):
        if self.token is None:
            return None

        token = self.token["id_token"].encode("utf8")

        kid = jwt.get_unverified_header(token)["kid"]
        public_key = None
        for key in self.jwks:
            if kid == key["kid"]:
                jwk = key
                break

        if jwk is None:
            if allow_refresh:
                logger.warn(
                    "could not find public key for id_token, "
                    "refreshing OIDC config"
                )
                cache.delete(CACHE_KEY_JWKS.format(self.site))
                cache.delete(CACHE_KEY_OPENID.format(self.site))

                return self.get_claims(allow_refesh=False)
            else:
                logger.warn("could not find public key for id_token")
                return None

        public_key = RSAAlgorithm.from_jwk(json.dumps(jwk))

        try:
            claims = jwt.decode(
                token,
                public_key,
                algoithm="RS256",
                audience=self.config.MICROSOFT_AUTH_CLIENT_ID,
            )
        except jwt.PyJWTError as e:
            logger.warn("could verify id_token sig: {}".format(e))
            return None

        return claims

    def authorization_url(self):
        """ Generates Microsoft/Xbox or a Office 365 Authorization URL """

        auth_url = self.openid_config["authorization_endpoint"]
        if self.config.MICROSOFT_AUTH_LOGIN_TYPE == LOGIN_TYPE_XBL:
            auth_url = self._xbox_authorization_url

        return super().authorization_url(auth_url, response_mode="form_post")

    def fetch_token(self, **kwargs):
        """ Fetchs OAuth2 Token with given kwargs"""

        return super().fetch_token(  # pragma: no cover
            self.openid_config["token_endpoint"],
            client_secret=self.config.MICROSOFT_AUTH_CLIENT_SECRET,
            **kwargs
        )

    def fetch_xbox_token(self):
        """ Fetches Xbox Live Auth token.

            token must contain a valid access_token
                - retrieved from fetch_token

            Reversed engineered from existing Github repos,
                no "official" API docs from Microsoft

            Response will be similar to
            {
                'Token': 'token',
                'IssueInstant': '2016-09-27T15:01:45.225637Z',
                'DisplayClaims': {'xui': [{'uhs': '###################'}]},
                'NotAfter': '2016-10-11T15:01:45.225637Z'
            }
        """

        # Content-type MUST be json for Xbox Live
        headers = {
            "Content-type": "application/json",
            "Accept": "application/json",
        }
        params = {
            "RelyingParty": "http://auth.xboxlive.com",
            "TokenType": "JWT",
            "Properties": {
                "AuthMethod": "RPS",
                "SiteName": "user.auth.xboxlive.com",
                "RpsTicket": "d={}".format(self.token["access_token"]),
            },
        }
        response = requests.post(
            self._xbox_token_url, data=json.dumps(params), headers=headers
        )

        if response.status_code == 200:
            self.xbox_token = response.json()

        return self.xbox_token

    def get_xbox_profile(self):
        """
            Fetches the Xbox Live user profile from Xbox servers

            xbox_token must contain a valid Xbox Live token
                - retrieved from fetch_xbox_token

            Reversed engineered from existing Github repos,
                no "official" API docs from Microsoft

            Response will be similar to
            {
                'NotAfter': '2016-09-28T07:19:21.9608601Z',
                'DisplayClaims': {
                    'xui': [
                        {
                            'agg': 'Adult',
                            'uhs': '###################',
                            'usr': '###',
                            'xid': '################',
                            'prv': '### ### ###...',
                            'gtg': 'Gamertag'}]},
                'IssueInstant': '2016-09-27T15:19:21.9608601Z',
                'Token': 'token'}
        """

        if "Token" in self.xbox_token:
            # Content-type MUST be json for Xbox Live
            headers = {
                "Content-type": "application/json",
                "Accept": "application/json",
            }
            params = {
                "RelyingParty": "http://xboxlive.com",
                "TokenType": "JWT",
                "Properties": {
                    "UserTokens": [self.xbox_token["Token"]],
                    "SandboxId": "RETAIL",
                },
            }
            response = requests.post(
                self._profile_url, data=json.dumps(params), headers=headers
            )

            if response.status_code == 200:
                return response.json()["DisplayClaims"]["xui"][0]
        return {}

    def valid_scopes(self, scopes):
        """ Validates response scopes based on MICROSOFT_AUTH_LOGIN_TYPE """

        scopes = set(scopes)
        required_scopes = None
        if self.config.MICROSOFT_AUTH_LOGIN_TYPE == LOGIN_TYPE_XBL:
            required_scopes = set(self.SCOPE_XBL)
        else:
            required_scopes = set(self.SCOPE_MICROSOFT)

        # verify all require_scopes are in scopes
        return required_scopes <= scopes



from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.test import override_settings, RequestFactory
from django.urls import reverse

from sites_microsoft_auth.admin import _register_admins
from sites_microsoft_auth.old_conf import LOGIN_TYPE_MA, LOGIN_TYPE_XBL
from sites_microsoft_auth.models import MicrosoftAccount, XboxLiveAccount

from . import TestCase


class AdminTests(TestCase):
    def setUp(self):
        super().setUp()

        User = get_user_model()
        self.factory = RequestFactory()
        self.request = self.factory.get("/")
        self.site = get_current_site(self.request)

        self.user = User.objects.create_superuser(
            "test", "test@example.com", "password1", site=self.site
        )
        self.microsoft_account = MicrosoftAccount.objects.create(
            microsoft_id="test", user=self.user, site=self.site
        )
        self.xbox_account = XboxLiveAccount.objects.create(
            xbox_id="test", gamertag="test", user=self.user, site=self.site
        )

        self.client.force_login(self.user)

    @override_settings(
        MICROSOFT_AUTH_LOGIN_TYPE=LOGIN_TYPE_MA,
        MICROSOFT_AUTH_REGISTER_INACTIVE_ADMIN=False,
    )
    def test_admin_classes_microsoft_auth(self):
        """ Verify only Microsoft Auth classes are injected """

        _register_admins()

        self.client.get(reverse("admin:index"))
        self.client.get(
            reverse("admin:{0}_{1}_change".format(self.user._meta.app_label, self.user._meta.model_name), args=(self.user.id,))
        )

        self.client.get(
            reverse(
                "admin:sites_microsoft_auth_microsoftaccount_change",
                args=(self.microsoft_account.id,),
            )
        )

    @override_settings(
        MICROSOFT_AUTH_LOGIN_TYPE=LOGIN_TYPE_XBL,
        MICROSOFT_AUTH_REGISTER_INACTIVE_ADMIN=False,
    )
    def test_admin_classes_xbox(self):
        """ Verify only Xbox classes are injected """

        _register_admins()

        self.client.get(reverse("admin:index"))
        self.client.get(
            reverse("admin:{0}_{1}_change".format(self.user._meta.app_label, self.user._meta.model_name), args=(self.user.id,))
        )

        self.client.get(
            reverse(
                "admin:sites_microsoft_auth_xboxliveaccount_change",
                args=(self.xbox_account.id,),
            )
        )

    @override_settings(MICROSOFT_AUTH_REGISTER_INACTIVE_ADMIN=True)
    def test_admin_classes_both(self):
        """ Verify both admin classes are injected """

        _register_admins()

        self.client.get(reverse("admin:index"))
        self.client.get(
            reverse("admin:{0}_{1}_change".format(self.user._meta.app_label, self.user._meta.model_name), args=(self.user.id,))
        )

        self.client.get(
            reverse(
                "admin:sites_microsoft_auth_microsoftaccount_change",
                args=(self.microsoft_account.id,),
            )
        )
        self.client.get(
            reverse(
                "admin:sites_microsoft_auth_xboxliveaccount_change",
                args=(self.xbox_account.id,),
            )
        )



import json
import urllib.parse
from unittest.mock import Mock, patch
from urllib.parse import parse_qs, urlparse

from django.contrib.sites.models import Site
from django.test import RequestFactory, override_settings

from sites_microsoft_auth.client import MicrosoftClient
from sites_microsoft_auth.conf import get_conf
from sites_microsoft_auth.old_conf import LOGIN_TYPE_XBL

from . import TestCase

STATE = "test_state"
CLIENT_ID = "test_client_id"
REDIRECT_URI = "https://testserver/microsoft/auth-callback/"
ACCESS_TOKEN = "test_access_token"
XBOX_TOKEN = "test_xbox_token"
XBOX_PROFILE = "test_profile"


@override_settings(SITE_ID=1)
class ClientTests(TestCase):
    @classmethod
    def setUpClass(self):
        super().setUpClass()

    def setUp(self):
        super().setUp()

        self.factory = RequestFactory()

    def _get_auth_url(self, base_url, scopes=MicrosoftClient.SCOPE_MICROSOFT):
        args = {
            "response_type": "code",
            "client_id": CLIENT_ID,
            "redirect_uri": REDIRECT_URI,
            "scope": " ".join(scopes),
            "state": STATE,
            "response_mode": "form_post",
        }
        return (base_url + "?" + urllib.parse.urlencode(args), STATE)

    def _assert_auth_url(self, expected, actual):
        # parse urls
        e_url = urlparse(expected[0])
        e_qs = parse_qs(e_url.query)
        a_url = urlparse(actual[0])
        a_qs = parse_qs(a_url.query)

        # assert url
        self.assertEqual(e_url.scheme, a_url.scheme)
        self.assertEqual(e_url.path, a_url.path)
        self.assertEqual(e_url.netloc, a_url.netloc)
        self.assertEqual(len(e_qs.items()), len(a_qs.items()))
        for key, value in e_qs.items():
            self.assertEqual(value, a_qs[key])

        # assert state
        self.assertEqual(expected[1], actual[1])

    def test_scope(self):
        expected_scopes = " ".join(MicrosoftClient.SCOPE_MICROSOFT)

        auth_client = MicrosoftClient()
        self.assertEqual(expected_scopes, auth_client.scope)

    def test_xbox_scopes(self):
        expected_scopes = " ".join(MicrosoftClient.SCOPE_XBL)
        config = get_conf(None)
        config.login_type = LOGIN_TYPE_XBL
        config.save()

        auth_client = MicrosoftClient()
        self.assertEqual(expected_scopes, auth_client.scope)

    def test_state(self):
        auth_client = MicrosoftClient(state=STATE)
        self.assertEqual(STATE, auth_client.state)

    def test_redirect_uri(self):
        auth_client = MicrosoftClient()
        self.assertEqual(REDIRECT_URI, auth_client.redirect_uri)

    def test_authorization_url(self):
        config = get_conf(None)
        config.client_id = CLIENT_ID
        config.save()

        auth_client = MicrosoftClient(state=STATE)

        base_url = auth_client.openid_config["authorization_endpoint"]
        expected_auth_url = self._get_auth_url(base_url)

        self._assert_auth_url(
            expected_auth_url, auth_client.authorization_url()
        )

    def test_authorization_url_with_xbl(self):
        base_url = MicrosoftClient._xbox_authorization_url
        expected_auth_url = self._get_auth_url(
            base_url, scopes=MicrosoftClient.SCOPE_XBL
        )

        config = get_conf(None)
        config.client_id = CLIENT_ID
        config.login_type = LOGIN_TYPE_XBL
        config.save()

        auth_client = MicrosoftClient(state=STATE)
        self._assert_auth_url(
            expected_auth_url, auth_client.authorization_url()
        )

    @patch("sites_microsoft_auth.client.requests")
    def test_fetch_xbox_token(self, mock_requests):
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = XBOX_TOKEN
        mock_requests.post.return_value = mock_response

        auth_client = MicrosoftClient()
        auth_client.token = {"access_token": ACCESS_TOKEN}
        xbox_token = auth_client.fetch_xbox_token()

        self.assertEqual(XBOX_TOKEN, xbox_token)
        self.assertEqual(XBOX_TOKEN, auth_client.xbox_token)

    @patch("sites_microsoft_auth.client.requests")
    def test_fetch_xbox_token_params(self, mock_requests):
        expected_headers = {
            "Content-type": "application/json",
            "Accept": "application/json",
        }
        expected_data = json.dumps(
            {
                "RelyingParty": "http://auth.xboxlive.com",
                "TokenType": "JWT",
                "Properties": {
                    "AuthMethod": "RPS",
                    "SiteName": "user.auth.xboxlive.com",
                    "RpsTicket": "d={}".format(ACCESS_TOKEN),
                },
            }
        )

        auth_client = MicrosoftClient()
        auth_client.token = {"access_token": ACCESS_TOKEN}
        auth_client.fetch_xbox_token()

        mock_requests.post.assert_called_with(
            MicrosoftClient._xbox_token_url,
            data=expected_data,
            headers=expected_headers,
        )

    @patch("sites_microsoft_auth.client.requests")
    def test_fetch_xbox_token_bad_response(self, mock_requests):
        mock_response = Mock()
        mock_response.status_code = 400
        mock_requests.post.return_value = mock_response

        auth_client = MicrosoftClient()
        auth_client.token = {"access_token": ACCESS_TOKEN}
        xbox_token = auth_client.fetch_xbox_token()

        self.assertEqual({}, xbox_token)
        self.assertEqual({}, auth_client.xbox_token)

    @patch("sites_microsoft_auth.client.requests")
    def test_get_xbox_profile(self, mock_requests):
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "DisplayClaims": {"xui": [XBOX_PROFILE]}
        }
        mock_requests.post.return_value = mock_response

        auth_client = MicrosoftClient()
        auth_client.xbox_token = {"Token": XBOX_TOKEN}
        xbox_profile = auth_client.get_xbox_profile()

        self.assertEqual(XBOX_PROFILE, xbox_profile)

    @patch("sites_microsoft_auth.client.requests")
    def test_get_xbox_profile_params(self, mock_requests):
        expected_headers = {
            "Content-type": "application/json",
            "Accept": "application/json",
        }
        expected_data = json.dumps(
            {
                "RelyingParty": "http://xboxlive.com",
                "TokenType": "JWT",
                "Properties": {
                    "UserTokens": [XBOX_TOKEN],
                    "SandboxId": "RETAIL",
                },
            }
        )

        auth_client = MicrosoftClient()
        auth_client.xbox_token = {"Token": XBOX_TOKEN}
        auth_client.get_xbox_profile()

        mock_requests.post.assert_called_with(
            MicrosoftClient._profile_url,
            data=expected_data,
            headers=expected_headers,
        )

    @patch("sites_microsoft_auth.client.requests")
    def test_get_xbox_profile_bad_response(self, mock_requests):
        mock_response = Mock()
        mock_response.status_code = 400
        mock_requests.post.return_value = mock_response

        auth_client = MicrosoftClient()
        auth_client.xbox_token = {"Token": XBOX_TOKEN}
        xbox_profile = auth_client.get_xbox_profile()

        self.assertEqual({}, xbox_profile)

    def test_get_xbox_profile_no_token(self):
        auth_client = MicrosoftClient()
        xbox_profile = auth_client.get_xbox_profile()

        self.assertEqual({}, xbox_profile)

    def test_valid_scopes(self):
        scopes = MicrosoftClient.SCOPE_MICROSOFT

        auth_client = MicrosoftClient()
        self.assertTrue(auth_client.valid_scopes(scopes))

    def test_valid_scopes_invalid(self):
        scopes = ["fake"]

        auth_client = MicrosoftClient()
        self.assertFalse(auth_client.valid_scopes(scopes))

    def test_valid_scopes_xbox(self):
        scopes = MicrosoftClient.SCOPE_XBL

        config = get_conf(None)
        config.login_type = LOGIN_TYPE_XBL
        config.save()

        auth_client = MicrosoftClient()
        self.assertTrue(auth_client.valid_scopes(scopes))

    @override_settings(
        SITE_ID=None, ALLOWED_HOSTS=["example.com", "testserver"]
    )
    def test_alternative_site(self):
        self.assertEqual(Site.objects.get(pk=1).domain, "testserver")

        Site.objects.create(domain="example.com", name="example.com")

        request = self.factory.get("/", HTTP_HOST="example.com")

        self.assertEqual(
            Site.objects.get_current(request).domain, "example.com"
        )

        client = MicrosoftClient(request=request)

        self.assertIn("example.com", client.authorization_url()[0])



from django.contrib.sites.shortcuts import get_current_site

from .models import SiteConfiguration


LOGIN_TYPE_MA = SiteConfiguration.LOGIN_TYPE_MA
LOGIN_TYPE_XBL = SiteConfiguration.LOGIN_TYPE_XBL
HOOK_SETTINGS = [
    "MICROSOFT_AUTH_AUTHENTICATE_HOOK",
    "MICROSOFT_AUTH_CALLBACK_HOOK",
]
CACHE_TIMEOUT = 86400
CACHE_KEY_OPENID = "microsoft_auth_openid_config_{site}"
CACHE_KEY_JWKS = "microsoft_auth_jwks_{site}"


def get_conf(request):
    site = get_current_site(request)
    conf, _ = SiteConfiguration.objects.get_or_create(site=site)
    return conf



# Generated by Django 2.2.5 on 2019-09-23 13:01

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('sites', '__latest__'),
    ]

    operations = [
        migrations.CreateModel(
            name='MicrosoftAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('microsoft_id', models.CharField(max_length=64, verbose_name='microsoft account id')),
            ],
        ),
        migrations.CreateModel(
            name='SiteConfiguration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('login_enabled', models.BooleanField(default=False)),
                ('login_type', models.CharField(choices=[('ma', 'Microsoft Account'), ('xbl', 'Xbox Live Account')], max_length=2)),
                ('tenant_id', models.CharField(default='common', max_length=80)),
                ('client_id', models.CharField(blank=True, default='', max_length=80)),
                ('client_secret', models.CharField(blank=True, default='', max_length=80)),
                ('extra_scopes', models.CharField(blank=True, default='', max_length=80)),
                ('auto_create', models.BooleanField(default=True)),
                ('register_inactive_admin', models.BooleanField(default=False)),
                ('xbl_sync_username', models.BooleanField(default=False)),
                ('auto_replace_accounts', models.BooleanField(default=False)),
                ('authenticate_hook', models.CharField(blank=True, default='', max_length=100)),
                ('callback_hook', models.CharField(blank=True, default='', max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='XboxLiveAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('xbox_id', models.CharField(max_length=32, unique=True, verbose_name='xbox user id')),
                ('gamertag', models.CharField(max_length=16, verbose_name='xbox live gamertag')),
            ],
        ),
    ]



import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

ALLOWED_HOSTS = ["*"]

SECRET_KEY = "fake-key"
INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "sites_user",
    "sites_microsoft_auth",
    "tests",
]

AUTH_USER_MODEL = 'sites_user.TestUser'

try:
    import djangoql  # noqa

    INSTALLED_APPS.append("djangoql")
except ImportError:
    pass

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}

ROOT_URLCONF = "tests.site.urls"

LANGUAGE_CODE = "en-us"

STATIC_URL = "/static/"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "site", "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

USE_I18N = True

USE_L10N = True

LANGUAGE_CODE = "en-us"

# this much be after the majority of your other settings
from sites_microsoft_auth.old_conf import DEFAULT_CONFIG  # isort:skip # noqa

CONSTANCE_BACKEND = "constance.backends.database.DatabaseBackend"
CONSTANCE_CONFIG = DEFAULT_CONFIG["defaults"]
CONSTANCE_CONFIG_FIELDSETS = DEFAULT_CONFIG["fieldsets"]
CONSTANCE_ADDITIONAL_FIELDS = DEFAULT_CONFIG["fields"]

# only use these settings when started via runserver manually
if "LOAD_DEV_SETTINGS" in os.environ:
    DEBUG = True
    TEMPLATES[0]["OPTIONS"]["context_processors"] += [
        "sites_microsoft_auth.context_processors.microsoft"
    ]

    AUTHENTICATION_BACKENDS = [
        "sites_microsoft_auth.backends.MicrosoftAuthenticationBackend",
        "django.contrib.auth.backends.ModelBackend",
    ]

    INSTALLED_APPS += ["constance", "constance.backends.database"]

    try:
        # At a minimum you still need to provide the following settings
        # MICROSOFT_AUTH_CLIENT_ID
        # MICROSOFT_AUTH_CLIENT_SECRET
        from .local import *  # noqa
    except ImportError:
        pass



"""Tests for `sites_microsoft_auth.conf` modules

Note: This file must be last in test runner as Constance does not get
cleaned correctly afterwards

isort:skip_file
"""

from unittest.mock import patch

import pytest
from django.core.management import call_command
from django.test import modify_settings

from sites_microsoft_auth.old_conf import DEFAULT_CONFIG

from . import TestCase
from .test_conf import ConfTests

pytest.importorskip("constance")


@modify_settings(
    INSTALLED_APPS={"prepend": ["constance", "constance.backends.database"]}
)
class ConstanceTests(TestCase):
    def test_constance_admin(self):
        self.client.get("/admin/constance/config/")

    def test_config_constance_not_found(self):
        from sites_microsoft_auth.old_conf import config, init_config

        init_config()

        with self.assertRaises(AttributeError):
            config.NOT_A_REAL_SETTING


@modify_settings(
    INSTALLED_APPS={"prepend": ["constance", "constance.backends.database"]}
)
class ConstanceConfTests(ConfTests):
    def setUp(self):
        super().setUp()

        call_command("migrate", "database", "zero")
        call_command("migrate", "database", "0001")


@patch("constance.base.settings.CONFIG", DEFAULT_CONFIG["defaults"])
@patch("constance.base.settings.CONFIG_FIELDSETS", DEFAULT_CONFIG["fieldsets"])
class ConstanceWithDefaultsConfTests(ConstanceConfTests):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from os import path

from setuptools import find_packages, setup

import versioneer


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))

    return [line for line in lineiter if line and not line.startswith("#")]


BASE_DIR = path.abspath(path.dirname(__file__))

with open(path.join(BASE_DIR, "README.rst")) as readme_file:
    readme = readme_file.read()

with open(path.join(BASE_DIR, "HISTORY.rst")) as history_file:
    history = history_file.read()

req_files = {
    "dev": "reqs/dev.in",
    "ql": "reqs/ql.in",
    "requirements": "reqs/requirements.in",
    "setup": "reqs/setup.in",
    "test": "reqs/test.in",
}

requirements = {}
for req, req_file in req_files.items():
    reqs = parse_requirements(req_file)
    requirements[req] = reqs


setup(
    name="django_sites_microsoft_auth",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description=(
        "Simple app to enable Microsoft Account, Office 365 "
        "Enterprise, Azure AD, and Xbox Live authentication as a "
        "Django authentication backend on a per-site basis."
    ),
    long_description=readme + "\n\n" + history,
    author="Griffin Skudder",
    author_email="griffin@skudder.co.nz",
    url="https://github.com/gskudder/django_sites_microsoft_auth",
    packages=find_packages(include=["sites_microsoft_auth", "sites_microsoft_auth.*"]),
    include_package_data=True,
    install_requires=requirements["requirements"],
    license="MIT license",
    zip_safe=False,
    keywords="django_sites_microsoft_auth",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
    ],
    test_suite="tests",
    tests_require=requirements["test"],
    setup_requires=requirements["setup"],
    extras_require={
        "dev": requirements["dev"],
        "ql": requirements["ql"],
        "test": requirements["test"],
    },
)



import importlib

from .conf import HOOK_SETTINGS
from django.conf import settings
from .old_conf import config as global_config


def get_scheme(request, config=None):
    if config is None:
        config = global_config

    scheme = "https"
    if settings.DEBUG and request is not None:
        if "HTTP_X_FORWARDED_PROTO" in request.META:
            scheme = request.META["HTTP_X_FORWARDED_PROTO"]
        else:
            scheme = request.scheme
    return scheme


def get_hook(name):
    if name in HOOK_SETTINGS:
        hook_setting = getattr(global_config, name)
        if hook_setting != "":
            module_path, function_name = hook_setting.rsplit(".", 1)
            module = importlib.import_module(module_path)
            function = getattr(module, function_name)

            return function
    return None



# Generated by Django 2.2.5 on 2019-09-23 15:35

from django.db import migrations


def forward(apps, schema_editor):
    SiteConfiguration = apps.get_model('sites_microsoft_auth', 'SiteConfiguration')
    Site = apps.get_model('sites', 'Site')
    for site in Site.objects.all():
        SiteConfiguration.objects.get_or_create(site=site)


class Migration(migrations.Migration):

    dependencies = [
        ('sites_microsoft_auth', '0005_auto_20190923_1514'),
    ]

    operations = [
        migrations.RunPython(forward, migrations.RunPython.noop)
    ]



from django.apps import AppConfig


class MicrosoftAuthConfig(AppConfig):
    name = "sites_microsoft_auth"
    verbose_name = "Microsoft Auth"

    def ready(self):
        import sites_microsoft_auth.signals



"""Tests for `sites_microsoft_auth.conf` modules without Constance

"""

from django.test import override_settings

from sites_microsoft_auth.old_conf import DEFAULT_CONFIG, SimpleConfig

from . import TransactionTestCase


class SimpleTestConfig(SimpleConfig):
    pass


class SimpleTestNoDefaultConfig:
    pass


no_default_test_conf = SimpleTestNoDefaultConfig()
test_conf = SimpleTestConfig()


class ConfTests(TransactionTestCase):
    def test_default_settings(self):
        """ Tests all the default settings are initialized correctly """

        from sites_microsoft_auth.old_conf import config

        for key, setting in DEFAULT_CONFIG["defaults"].items():
            actual_setting = getattr(config, key)
            self.assertEqual(actual_setting, setting[0])

    def test_config_class(self):
        """ Tests MICROSOFT_AUTH_CONFIG_CLASS=None """

        from sites_microsoft_auth.old_conf import config

        self.assertTrue(isinstance(config, SimpleConfig))
        self.assertFalse(isinstance(config, SimpleTestConfig))

    @override_settings(MICROSOFT_AUTH_CONFIG_CLASS=None)
    def test_config_class_as_none(self):
        """ Tests MICROSOFT_AUTH_CONFIG_CLASS=None """

        from sites_microsoft_auth.old_conf import config

        self.assertTrue(isinstance(config, SimpleConfig))
        self.assertFalse(isinstance(config, SimpleTestConfig))

    @override_settings(MICROSOFT_AUTH_CONFIG_CLASS="tests.test_conf.test_conf")
    def test_custom_config_class(self):
        """ Tests MICROSOFT_AUTH_CONFIG_CLASS set to another class """
        from sites_microsoft_auth.old_conf import config

        self.assertTrue(isinstance(config, SimpleTestConfig))

    @override_settings(
        MICROSOFT_AUTH_CONFIG_CLASS="tests.test_conf.no_default_test_conf"
    )
    def test_custom_config_class_with_no_default(self):
        """ Tests MICROSOFT_AUTH_CONFIG_CLASS set to another class with no
            add_default_config option """
        from sites_microsoft_auth.old_conf import config

        self.assertTrue(isinstance(config, SimpleTestNoDefaultConfig))



from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site

from sites_microsoft_auth.models import MicrosoftAccount, XboxLiveAccount

from . import TestCase

USER_ID = "test_user_id"


class ModelsTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.site = Site.objects.get(id=1)

    def test_microsoft_account_str(self):
        a = MicrosoftAccount(microsoft_id=USER_ID, site=self.site)
        a.save()

        self.assertEqual(USER_ID, str(a))

    def test_xbox_account_str(self):
        a = XboxLiveAccount(gamertag=USER_ID, site=self.site)
        a.save()

        self.assertEqual(USER_ID, str(a))

    def test_username_with_spaces(self):
        User = get_user_model()

        u = User(username="Test username", site=self.site)
        u.set_unusable_password()
        u.full_clean()
